/* tudo.c V2.0 */
/* Programa para trabalhar com prototipo do 8031. Recursos: 		*/
/* --> transmitir repetitivamente uma mensagem pela porta serial  */
/* --> mostrar o que chega pela porta serial     						*/
/* --> fazer o boot do prototipo                 						*/
/* --> controlar o single step                   						*/

# include <stdio.h>
# include <stdlib.h>
# include <dos.h>
# include <conio.h>
# include <ctype.h>
# include <bios.h>
# include <string.h>
// # include <graphics.h>
# include <time.h>

# define TRUE			1
# define FALSE			0
# define K				1024
# define TAM_FILA		2*K			/* tamanho da fila circular					*/
# define B28800		0x4			// baud rate de 28.800
# define B19200		0x6			// baud rate de 19.200
# define B14400		0x8			// baud rate de 14.400
# define B9600			0xc			// baud rate de 9600
# define B4800			0x18			// baud rate de 4800
# define B2400			0x30			// baud rate de 2400

# define SET9600		(0x3|0x4|0x00|0xE0) 	/* 8 bits, 2 stop, no parity 		*/
# define SET4800		(0x3|0x4|0x00|0xC0) 	/* 8 bits, 2 stop, no parity 		*/
# define SET2400		(0x3|0x4|0x00|0xA0) 	/* 8 bits, 2 stop, no parity 		*/
# define SET1200		(0x3|0x4|0x00|0x80) 	/* 8 bits, 2 stop, no parity 		*/


/* definicoes para a gravacao de EPROM													*/
# define CMD_APAGADO			'\A'			/* ver se a EPROM esta apagada		*/
# define CMD_LEITURA			'\L'			/* comandar ckt para ler EPROM		*/
# define CMD_GRAV				'\G'			/* comando para iniciar gravacao		*/
# define CMD_SEG				'\S'			/* gravar o bit de seguranca        */
# define CMD_RECORD			'\R'			/* transmitir um outro record			*/
# define CMD_CONF				'\C'			/* conferir conteudo da EPROM			*/
# define ST_PROG				1				/* usar gravacao em modo standart	*/
# define QUICK_PROG			2				/* usar o QUICK PROGRAMMING PULSE	*/
# define TAM_4K				4				/* EPROM DE 4 KB							*/
# define TAM_8K				8				/* EPROM de 8 KB							*/
# define TIME_OUT_SEG		5				/* time out para retirar da fila		*/
# define DELAY					30				/* delay para enviar bytes para gravar	*/

/* enderecos dos diversos dispositivos 												*/
# define COM01			1
# define COM02			2
# define END_COM1		0x3f8			/* endereco da 8250 ligado na COM1			*/
# define END_COM2		0x2f8			/* endereco da 8250 ligado na COM2			*/
# define RXBIT			1				/* bit que indica chegada de um dado		*/
# define TXBIT 		0x20			/* bit que indica transmissao de um dado	*/
# define INT_COM1		0x10			/* bit para habilitar interrupcao COM1		*/
# define INT_COM2		0x8			/* bit para habilitar interrupcao COM2		*/
# define END_8259		0x20			/* endereco da 8259								*/
# define HAB_INTR		(END_8259+1)/* registrado do 8259 que habilita intr	*/


/* offset para acessar os registradores da 8250										*/
# define TX_BUF		0				/* TX buffer (DLAB=0)							*/
# define RX_BUF		0				/* RX buffer (DLAB=0)							*/
# define DLL			0				/* divisor latch LSB (DLAB=1)					*/
# define DLM			1				/* divisor latch MSB (DLAB=1)					*/
# define IER			1				/* interrupt eneble register					*/
# define IIR			2				/* interrupt identification register		*/
# define LCR			3				/* line control register						*/
# define MCR			4				/* modem control register						*/
# define LSR			5				/* line status register							*/
# define MSR			6				/* modem status register						*/

/* comandos validos	*/
# define PASSO			1
# define DUMP			2
# define TROCA			3
# define RESET			4
# define FIM			5
# define HELP			6

/* offset dos registradores dentro do vetor reg										*/
# define R0				0
# define PCON			1
# define SCON			2
# define TCON  		3
# define TMOD			4
# define TH1 			5
# define TL1			6
# define DPH 			7
# define DPL 			8
# define B 				9
# define A 				10
# define PSW 			11
# define PCH 			12
# define PCL   		13
# define PC_0 			14
# define PC_1 			15
# define PC_2			16
# define PC_3 			17
# define PC_4 			18
# define PC_5			19
# define PC_6			20
# define PC_7 			21
# define SP  			22
# define IP    		23
# define IE  			24
# define SBUF			25
# define TH0			26
# define TL0 			27
# define R1  			28
# define R2				29
# define R3  			30
# define R4				31
# define R5  			32
# define R6				33
# define R7  			34
# define TOTAL_REG	R7
# define TOTAL_RAM	127

/* flags do PSW da 8031																		*/
# define CY				0x80
# define AC				0x40
# define F0				0x20
# define RS1			0x10
# define RS0			0x8
# define OV				0x4
# define P				0x1

/* codigos das teclas especiais															*/
# define ESC			0x1b
# define CR				0xd
# define LF				0xa
# define SPACE			0x20

/* caracteres graficos para moldura 	*/
# define SUP_ESQ	0xc9
# define SUP_DIR	0xbb
# define INF_ESQ	0xc8
# define INF_DIR	0xbc
# define HOR    	0xcd
# define VER		0xba

/* posicao da janela	*/
# define XWI	10
# define YWI	5
# define XWF	60
# define YWF	20

/* funcoes que serao usadas																*/
void  interrupt(*antigo)(void);		/* guardar interrupcao antiga					*/
int 	poe (unsigned char byte);  /* colocar byte na fila circular				*/
int 	tira(unsigned char *byte);	/* retirar byte da fila circular				*/
int 	le_opc_num (int max);		/* ler opcoes numericas							*/
int 	le_opc(char *op);				/* ler opcoes com letras						*/
void 	troca_serial(void);
void 	sel_baud_rate(void);
void 	tx_serial(void);
void 	tx_rx_serial(void);
void 	rx_serial(void);
void 	boot(void);
void 	passo_passo(void);
int  	getfuncao(void);				/* ler teclas F1 a F10 e ESC					*/
void 	passo(void);
void 	dump(void);
void 	troca(void);
void 	reset(void);
void 	help(void);
void 	rec_reg(void);
void 	display_reg(void);
int  	le_comando(int *n_passo, int *faz_dump);
int  	get_hexa(int *dado);			/* ler valores em hexa, de 0 a 255 			*/
int  	disassembly(void);
int  	qual_bit(unsigned char bit);
int  	qual_reg (unsigned char rg);
int  	qual_rn  (unsigned char rn);
int  	relativo (unsigned int pc, unsigned char rel);
int  	faz_moldura(int xi, int yi, int xf, int yf, char *msg);
int  	guarda_tela(int xi, int yi, int xf, int yf);
int  	recupera_tela(int xi, int yi, int xf, int yf);
int  	le_opc_help(char *op);
int 	qq_tecla_help(void);
void 	get_enter(void);
void 	get_enter_silencio(void);
void	prep_serial(void);
void	prep_int(void);
void	desprep_int(void);

int	pprep_serial(void);
int 	des_intr(void);
int 	rec_msg(void);
int	sai_serial(int dado);
void	tx_rep(void);

/* variaveis globais																			*/
unsigned char fila[TAM_FILA];		/* fila circular para receber dados			*/
int pin,pout;							/* ponteiros para por e retirar da fila	*/
long contador;							/* contar bytes									*/
int porta,end_porta;					/* qual porta serial e seu endereco			*/
int baud_rate;							/* baud rate selecionado	               */
int hab_intr;							/* qual interrupcao vai habilitar			*/
unsigned char reg[TOTAL_REG+2];	/* receber registradores do 8031				*/
unsigned char ram[TOTAL_RAM+2];	/* receber ram interna							*/
char instr[20];
char *texto=" <<< HELP >>> ";
unsigned char far *tela;
unsigned char buffer[80*25];

/* rotina para receber dados pela porta serial										*/
void interrupt serial(void){
	contador++;
	poe(inportb(end_porta));
	outportb(0x20,0x20);
}

main(void){
	int opc;
	tela=MK_FP(0xb000,0x08000);
	pin=1;
	pout=0;
	porta=COM01;
	end_porta=END_COM1;
	baud_rate=28800;
	prep_serial();
	while (TRUE){
		clrscr();
		printf("         <<< SISTEMA DE DESENVOLVIMENTO PARA MCS51 (V2.0) >>>\n\n");
		printf("Preparado para COM%1d: %d, 8 dados, 2 STOP, sem paridade\n",porta,baud_rate);
		printf("\n\nAs opcoes existentes sao:\n\n");
		printf("1 - Boot do prototipo 8031.\n");
		printf("2 - Single step.\n");
		printf("3 - Enviar/Receber pela porta serial.\n");
		printf("4 - Enviar mensagem pela porta serial.\n");
		printf("5 - Receber pela porta serial.\n");
		printf("6 - Trocar de porta serial.\n");
		printf("7 - Selecionar outro Baud Rate.\n");
		printf("8 - Transmitir repetitivamente pela serial.\n");
		printf("9 - Retornar ao DOS.\n");
		printf("\nEscolha uma das opcoes : ");
		opc=le_opc_num(9);
		if (opc==9)	break;
		switch(opc){
			case 1: boot();				break;
			case 2: passo_passo();		break;
			case 3: tx_rx_serial();		break;
			case 4: tx_serial();			break;
			case 5: rx_serial();			break;
			case 6: troca_serial();		break;
			case 7: sel_baud_rate();	break;
			case 8: tx_rep();				break;
		}
	}
	printf("Terminar programa !");
	return(0);
}

/* (opc=8) transmitir repetitivamente pela serial	*/
void tx_rep(void){

	clrscr();
	printf("(COM%d, %d)       <<< TRANSMISSAO REPETITIVA PELO CANAL SERIAL >>>\n\n\n",porta,baud_rate);

	printf("Sera transmitido a sequencia 0x55 e 0xAA.\n");
	get_enter();
	printf("\n(COM%d, %d)Inicando a transmissao. Qualquer tecla interrompe.\n\n",porta,baud_rate);
	while (~kbhit()){
		outportb(end_porta+TX_BUF,0x55);
		while ( (inportb(end_porta+LSR)&TXBIT) == 0)	;
		outportb(end_porta+TX_BUF,0xAA);
		while ( (inportb(end_porta+LSR)&TXBIT) == 0)	;
	}
	printf("\n\nFim da transmissao.\n");
	printf("Qualquer tecla para retornar ao menu principal.");
	getch();
	getch();
}

/* (opc==1) fazer o boot da placa pela porta serial	*/
void boot(void){
	FILE *arq;
	char nome[50];
	unsigned char byte;
	int x,y,dado;
	long cont;

	clrscr();
	printf("(COM%d, %d)        <<< BOOT SERIAL >>>\n\n\n",porta,baud_rate);
	pin=1;
	pout=0;
	prep_int();		//habilitar interrupcao

	printf("\n\nAguardando chegada de uma possivel mensagem serial\n");
	printf("Qualquer tecla para seguir adiante\n\n");

	while(~kbhit()){
		if (tira(&byte)==TRUE)	printf("%c",byte);
	}
	outportb(end_porta+IER,0);							/* RX nao gera interrupcao	*/
	desprep_int();

	printf("\n\n\n");
	printf("<<< Transmitir o programa pelo canal serial (INTEL.HEX) >>>\n\n");
	printf("Qual o nome do arquivo ? ");
	scanf("%s",nome);
	arq=fopen(nome,"rb");
	while (arq==NULL){
		printf("Nao foi possivel abrir o arquivo: %s\n",nome);
		printf("Tente de novo. Qual o nome do arquivo ? ");
		scanf("%s",nome);
		arq=fopen(nome,"rb");
	}

	printf("Tudo pronto !  Tecle <ENTER> para iniciar a transmissao\n");
	get_enter();

	printf("Inicando a transmissao. Qualquer tecla interrompe.\n");
	cont=0;
	while(TRUE){
		dado=getc(arq);
		if (dado==EOF)	break;
		outportb(end_porta+TX_BUF,dado);
		cont++;
		putchar(dado);
		while ((inportb(end_porta+LSR)&TXBIT)==0)	;
	}
	printf("\n\nFim da transmissao. Transmitidos [%ld] bytes.\n\n",cont);
	printf("Qualquer tecla para voltar ao menu principal\n");
	getch();
}

/* (opc==2) realizar o controle do SINGLE STEP	*/
void passo_passo(void){
	int n_passo,faz_dump,cmd;
	clrscr();
	printf("(COM%d, %d)         <<< SINGLE STEP >>>\n\n\n",porta,baud_rate);
	printf("Aguardando 8031 enviar os registradores.\n");
	get_enter();

	/* fazer as inicializacoes para porta serial e interrupcao */
	prep_int();
	pin=1;
	pout=0;
	n_passo=0;
	faz_dump=FALSE;
	rec_reg();
	display_reg();
	while(TRUE){
		cmd=le_comando(&n_passo,&faz_dump);
/*		getchar();     */
		switch (cmd){
			case PASSO :	break;
			case DUMP  :	dump();	break;
			case TROCA :	troca();	break;
			case RESET :	reset();	break;
			case HELP  :   help();	break;
			case FIM	  : 	break;
		}

		if (cmd==FIM)	break;

		while (n_passo!=0){
			passo();
			if (faz_dump)	dump();
			n_passo--;
		}
	}
	/* desligar interrupcoes	*/
	desprep_int();
	printf("Terminado PASSO A PASSO.\n");
	printf("<ENTER> para retornar ao menu principal\n");
	get_enter();
}

/* (opc==3) Transmitir e receber pela porta serial		*/
void tx_rx_serial(void){
;
}


/* (opc==4 transmitir string ou arquivo pelo canal serial				*/
void tx_serial(void){
	FILE *arq;
	char nome[50];
	int x,y,dado,opc;
	long cont=0;

	clrscr();
	printf("(COM%d, %d)       <<< TRANSMISSAO PELO CANAL SERIAL >>>\n\n\n",porta,baud_rate);

	printf("Deseja transmitir um arquivo (Sim/Nao) ? ");
	opc=le_opc("SsNn");
	if (opc==1 || opc==2){
		printf("Qual o nome do arquivo ? ");
		scanf("%s",nome);
		arq=fopen(nome,"rb");
		while (arq==NULL){
			printf("Nao foi possivel abrir o arquivo: %s\n",nome);
			printf("Tente de novo. Qual o nome do arquivo ? ");
			scanf("%s",nome);
			arq=fopen(nome,"rb");
		}
		printf("Tudo pronto !  Tecle <ENTER> para iniciar a transmissao\n");
		get_enter();


		printf("(COM%d)Inicando a transmissao. Qualquer tecla interrompe.\n\n",porta);
		cont=0;
		while (~kbhit()){
			if ( (inportb(end_porta+LSR)&TXBIT) == TXBIT){
				dado=getc(arq);
				if (dado==EOF)	break;
				cont++;
				outportb(end_porta+TX_BUF,dado);
				printf("%c",dado);
				x=wherex();
				y=wherey();
				gotoxy(64,1);
				printf("  [%09ld]  ",cont);
				gotoxy(x,y);
			}
		}
	}
	else{
		printf("Tudo que for teclado sera transmitido. ESCAPE termina.\n");
		printf("[RETURN] transmite apenas 0xd (mudei e agora envia 0xa)\n\n");
		while(TRUE){
			dado=getch();
			if (dado==ESC)	break;
			cont++;
			outportb(end_porta+TX_BUF,dado);
			if (dado!=0xd)	putchar(dado);
			else{
				printf("\n");
				while( (inportb(end_porta+LSR)&0x20) == 0);
				outportb(end_porta+TX_BUF,0xa);
			}
			x=wherex();
			y=wherey();
			gotoxy(64,1);
			printf("  [%09ld]  ",cont);
			gotoxy(x,y);
		}
	}

	printf("\n\nFim da transmissao. Transmitidos %ld bytes.\n",cont);
	printf("Qualquer tecla para retornar ao menu principal ");
	getch();

}


/* (opc==5) receber dados pelo canal serial, pode gravar em arquivo	*/
void rx_serial(void)
{
	FILE	*arq;
	char nome[50];
	int x,y,opc,tela,arquivo;
	unsigned char byte;

	tela=TRUE;
	arquivo=FALSE;
	clrscr();
	printf("(COM%d, %d)          <<< RECEBER DADOS PELO CANAL SERIAL >>>\n\n\n",porta,baud_rate);
	printf("Deseja guardar em arquivo (Sim/Nao) ? ");
	opc=le_opc("SsNn");
	if (opc==1 || opc==2){
		arquivo=TRUE;
		printf("Qual o nome do arquivo ? ");
		scanf("%s",nome);
		arq=fopen(nome,"wb");
		while(arq==NULL){
			printf("Nao foi possivel abrir [%s].\n",nome);
			printf("Tente novamente. Qual o nome do arquivo ? ");
			scanf("%s",nome);
			arq=fopen(nome,"wb");
		}
		printf("Para grandes quantidades de dados e conveniente nao exibir na tela\n");
		printf("Deseja suprimir exibicao na tela (Sim/Nao) : ");
		opc=le_opc("SsNn");
		if (opc==1 || opc==2)	tela=FALSE;
	}
	printf("Tudo pronto. Tecle <ENTER> para iniciar recepcao\n");
	get_enter();
	prep_int();	// habilitar interrupcao
	printf("\n\n(COM%d)  Iniciando a recepcao. Qualquer tecla interrompe.\n",porta+1);

	contador=0;
	while(~kbhit()){
		if (tira(&byte)){
			if (tela){
				printf("%c",byte);
				x=wherex();
				y=wherey();
				gotoxy(64,1);
				printf("  [%09ld]  ",contador);
				gotoxy(x,y);
			}
			if (arquivo){
				putc(byte, arq);
				if (contador%100 == 0) {
					x=wherex();
					y=wherey();
					gotoxy(64,1);
					printf("  [%09ld]  ",contador);
					gotoxy(x,y);
				}
			}
		}
	}

	desprep_int();		//desabilitar interrupcao

	getch();
	printf("\n\n");
	if (arquivo){
		printf("Esvaziando o buffer de recepcao\n");
		while(tira(&byte))	putc(byte,arq);
		printf("Fechando o arquivo.\n");
		fclose(arq);
	}
	printf("\n\nRecebidos %ld bytes\n\n",contador);
	printf("Qualquer tecla para retornar ao menu principal ");
	getch();
}


/* (opc==6) trocar porta serial												*/
void troca_serial(void){
	int opc;
	clrscr();
	printf("             <<< TROCAR PORTA SERIAL >>>\n\n\n");
	printf("Preparado para COM%1d: %d, 8 dados, 2 STOP, sem paridade\n\n",porta,baud_rate);

	printf("Portas disponiveis:\n");
	printf("1 - COM1\n");
	printf("2 - COM2\n");
	printf("3 - Retornar\n");
	printf("Qual porta serial deseja usar: ");
	opc=le_opc_num(3);
	if (opc!=3){
		switch(opc){
			case 1: porta=COM01;	end_porta=END_COM1;	break;
			case 2: porta=COM02;	end_porta=END_COM2;	break;
		}
		prep_serial();
		printf("\n\nPreparado para COM%d: %d, 8 dados, 2 STOP, sem paridade\n",porta,baud_rate);
		printf("Qualquer tecla para retornar ao menu principal ");
		getch();
	}
}

/* (opc==7) selecionao o Baud Rate												*/
void sel_baud_rate(void){
	int opc;
	clrscr();
	printf("             <<< SELECIONAR O BAUD RATE >>>\n\n\n");
	printf("Preparado para COM%1d: %d, 8 dados, 2 STOP, sem paridade\n\n",porta,baud_rate);

	printf("BAUD RATE disponiveis:\n");
	printf("1 -  2400 bauds (bps)\n");
	printf("2 -  4800 bauds (bps)\n");
	printf("3 -  9600 bauds (bps)\n");
	printf("4 - 14400 bauds (bps)\n");
	printf("5 - 19200 bauds (bps)\n");
	printf("6 - 28800 bauds (bps)\n");
	printf("7 - Retornar\n");
	printf("Qual BAUD RATE deseja usar: ");
	opc=le_opc_num(7);
	if (opc!=7){
		switch(opc){
			case 1: baud_rate= 2400;	break;
			case 2: baud_rate= 4800;	break;
			case 3: baud_rate= 9600;	break;
			case 4: baud_rate=14400;	break;
			case 5: baud_rate=19200;	break;
			case 6: baud_rate=28800;	break;
			default:baud_rate= 9600;	break;
		}
		prep_serial();
		printf("\n\nPreparado para COM%d: %d, 8 dados, 2 STOP, sem paridade\n",porta,baud_rate);
		printf("Qualquer tecla para retornar ao menu principal ");
		getch();
	}
}

/* Preparar a porta serial	*/
void prep_serial(void){
	outportb(end_porta+LCR,0x87);		//DLAB=1
	switch(baud_rate){
			case 2400:	outportb(end_porta+DLL,B2400);
							outportb(end_porta+DLM,B2400>>8);	break;
			case 4800:	outportb(end_porta+DLL,B4800);
							outportb(end_porta+DLM,B4800>>8);	break;
			case 9600:	outportb(end_porta+DLL,B9600);
							outportb(end_porta+DLM,B9600>>8);	break;
			case 14400:	outportb(end_porta+DLL,B14400);
							outportb(end_porta+DLM,B14400>>8);	break;
			case 19200:	outportb(end_porta+DLL,B19200);
							outportb(end_porta+DLM,B19200>>8);	break;
			case 28800:	outportb(end_porta+DLL,B28800);
							outportb(end_porta+DLM,B28800>>8);	break;
			default:		outportb(end_porta+DLL,B9600);
							outportb(end_porta+DLM,B9600>>8);	break;
	}
	outportb(end_porta+LCR,0x7);		//DLAB=1, 8 bits, sem par., 2 stop
	outportb(end_porta+IER,0);			//sem interrupcao
}


// Preparar para usar interrupcao na recepcao serial
void prep_int(void){
	inport(end_porta+RX_BUF);		//leitura para esvaziar buffer recep.
	outportb(end_porta+MCR,8);
	outportb(end_porta+IER,1);
	if (porta==COM01){
		antigo=getvect(0xc);
		setvect(0xc,serial);
		outportb(HAB_INTR,inportb(HAB_INTR)&~INT_COM1);
	}
	else{
		antigo=getvect(0xb);
		setvect(0xb,serial);
		outportb(HAB_INTR,inportb(HAB_INTR)&~INT_COM2);
	}
	outportb(END_8259,0x20);
}

// Liberar a interrupcao usada na recepcao serial
void desprep_int(void){
	outportb(end_porta+IER,0);
	if (porta==COM01){
		setvect(0xc,antigo);
		outportb(HAB_INTR,inportb(HAB_INTR)|INT_COM1);
	}
	else{
		setvect(0xb,antigo);
		outportb(HAB_INTR,inportb(HAB_INTR)|INT_COM2);
	}
}


/* executar PASSO */
void passo(void)
{
	outportb(end_porta+TX_BUF,'P');
	while( (inportb(end_porta+LSR)&TXBIT) != TXBIT)	;
	rec_reg();
	display_reg();
}

/* executar DUMP */
void dump(void)
{
	int i,j;
	unsigned char byte;
	outportb(end_porta+TX_BUF,'D');
	while( (inportb(end_porta+LSR)&TXBIT) != TXBIT)	;

	/* receber RAM interna	*/
	for (i=TOTAL_RAM; i>=0; i--){
		while (tira(&byte)==FALSE)	;
		ram[i]=byte;
	}


	/* exibir a RAM interna	*/
	printf("       ");
	printf("0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F       <----ASCII---->\n");
	for (i=0; i<TOTAL_RAM; i+=16){
		printf("<%02X>   ",i);
		for (j=0; j<16; j++)		printf("%02X ",ram[i+j]);
		printf("     ");
		for (j=0; j<16; j++){
			if (ram[i+j]> 0x20 && ram[i+j]< 0xff)	printf("%c",ram[i+j]);
			else														printf(".");
		}
		printf("\n");
	}
}

/* executar RESET */
void reset(void)
{
	unsigned char byte;
	contador=0;
	outportb(end_porta+TX_BUF,'R');
	while( (inportb(end_porta+LSR)&TXBIT) != TXBIT)	;
	printf("\nMensagem recebida: ");
	byte=0;
	while (byte!=0xa){
		while (tira(&byte)==FALSE)	;
		printf("%c",byte);
	}
	rec_reg();
	display_reg();
}


/* executar uma TROCA */
void troca(void)
{
	int adr,dado,retorno;
	retorno=FALSE;
	while(!retorno){
		printf("\nQual o endereco, em HEXA, que deve ser alterado (0 ... 0FFH) ? ");
		retorno=get_hexa(&adr);
	}
	retorno=FALSE;
	while(!retorno){
		printf("Qual o dado, em HEXA, que deve ser escrito (0 ... 0FFH) ? ");
		retorno=get_hexa(&dado);
	}
	printf("Escrever 0%02XH(%03dD) no endereco 0%02XH(%03dD).\n",dado,dado,adr,adr);
	outportb(end_porta+TX_BUF,'T');
	while( (inportb(end_porta+LSR)&TXBIT) != TXBIT)	;
	outportb(end_porta+TX_BUF,adr);
	while( (inportb(end_porta+LSR)&TXBIT) != TXBIT)	;
	outportb(end_porta+TX_BUF,dado);
	while( (inportb(end_porta+LSR)&TXBIT) != TXBIT)	;
}

/* rotina para receber os registradores do 8031	*/
void rec_reg(void)
{
	int i;
	unsigned char byte;
/*	printf("\nReceber os registradores\n");   */

	for (i=0; i<=TOTAL_REG; i++){
		while (tira(&byte)==FALSE)		;

		reg[i]=byte;
/*		printf("(%03d,%02X)  ",i,byte)   */;
	}
}

/* rotina para mostrar os registradores na tela	*/
void display_reg(void)
{
	int banco;
	disassembly();
	strupr(instr);
	printf("\n");

	/* linha 1 */
	printf("<%s>    ",instr);
	gotoxy(19,gety());
	printf("PC=%02X%02X   [%02X %02X %02X %02X]   (",reg[PCH],reg[PCL],reg[PC_0],reg[PC_1],reg[PC_2],reg[PC_3]);
	if ( (reg[PSW]&CY)==CY )	printf("CY ");
	else								printf("cy ");
	if ( (reg[PSW]&AC)==AC )	printf("AC ");
	else								printf("ac ");
	if ( (reg[PSW]&OV)==OV )	printf("OV ");
	else								printf("ov ");
	if ( (reg[PSW]&F0)==F0 )	printf("F0");
	else								printf("f0");
	banco= ((reg[PSW]&RS1) >> 3) + ((reg[PSW]&RS0) >> 3);
	printf(") BK=%02d  R0=%02X  R1=%02X\n",banco,reg[R0],reg[R1]);

	/* linha 2 */
	printf(" A  B PSW SP DPTR IP IE TCON TMOD SCON SBUF PCON TIM0 TIM1 R2 R3 R4 R5 R6 R7\n");
	/*        A    B   PSW   SP  */
	printf("%02X %02X %02X  %02X ",reg[A],reg[B],reg[PSW],reg[SP]);
	/*        DPTR    IP   IE	*/
	printf("%02X%02X %02X %02X  ",reg[DPH],reg[DPL],reg[IP],reg[IE]);
	/*      TCON   TMOD   SCON   SBUF  */
	printf("%02X   %02X   %02X   %02X   ",reg[TCON],reg[TMOD],reg[SCON],reg[SBUF]);
	/*      PCON     TIM0    TIM1	*/
	printf("%02X  %02X%02X %02X%02X ",reg[PCON],reg[TH0],reg[TL0],reg[TH1],reg[TL1]);
	/*       R2    R3   R4	*/
	printf("%02X %02X %02X ",reg[R2],reg[R3],reg[R4]);
	/*       R5    R6   R7	*/
	printf("%02X %02X %02X ",reg[R5],reg[R6],reg[R7]);
	printf("\n");

}

/* rotina para ler os comandos validos	*/
le_comando(int *n_passo, int *faz_dump)
{
	char	c[20];
	while(TRUE){
		printf("\nP - passo,  D - dump RAM, T - troca,  R - reset,  H - help,  X - fim\n");
		printf("Qual o comando (P, D, R, T, Pn, PD, H, X) ? ");
		scanf("%s",c);
		strupr(c);
		*n_passo=0;
		if 	  (c[0]=='D')	return(DUMP);
		else if (c[0]=='T')	return(TROCA);
		else if (c[0]=='R')	return(RESET);
		else if (c[0]=='X')	return(FIM);
		else if (c[0]=='H')	return(HELP);
		else if (c[0]=='P'){
			if (c[1]=='\0'){
				*n_passo=1;
				return(PASSO);
			}
			else if (c[1]=='D' &&  c[2]=='\0'){
				*n_passo=1;
				if (*faz_dump)	*faz_dump=FALSE;
				else				*faz_dump=TRUE;
				return(PASSO);
			}
			else{
				*n_passo=atoi(c+1);
				if (*n_passo!=0)	return(PASSO);
			}
		}
		printf("Nao entendi o comando. Tente novamente\n");
	}
	return(0);
}

/* funcao para ler valores em hexa	*/
get_hexa(int *dado)
{
	char c[20];
	scanf("%s",c);
	strupr(c);
	c[0]-='0';
	if (c[0]>9){
		c[0]-=7;
		if (c[0]>15 || c[0]<0)	return(FALSE);
	}
	if (c[1]=='\0'){
		*dado=c[0];
		return(TRUE);
	}
	c[1]-='0';
	if (c[1]>9){
		c[1]-=7;
		if (c[1]>15 || c[1]<0)	return(FALSE);
	}
	if (c[2]=='\0'){
		*dado=c[0]*16+c[1];
		return(TRUE);
	}
	else	return(FALSE);
}

/* Esperar que saia um byte da fila, existe um time out	*/
int tira_byte_fila(void)
{
	unsigned char byte;
	time_t tv,tn;
	int x,y,z=0;
	x=wherex();
	y=wherey();
	tv=time(NULL);
	while(tira(&byte)==FALSE){
		gotoxy(x,y);
		switch(z++){
			case 0: printf("|");		break;
			case 1: printf("/"); 	break;
			case 2: printf("-"); 	break;
			case 3: printf("\\");	break;
			case 4: z=0;            break;
		}
		tn=time(NULL);
		if ((tn-tv)>TIME_OUT_SEG){
			gotoxy(x,y);
			return(EOF);
		}
	}
	return(byte);
}

/* rotina para retirar um byte na fila circular				*/
/* Retorna: FALSE ==> se nao conseguiu retirar da fila	*/
/*          TRUE  ==> se     conseguiu retirar da fila	*/
int tira( unsigned char *byte)
{
	int temp;
	if ( (temp=pout+1)==TAM_FILA) temp=0;
	if (temp==pin)		return(FALSE);
	pout=temp;
	*byte=fila[pout];
	return(TRUE);
}

/* rotina para colocar um byte na fila circular				*/
/* Retorna: FALSE ==> se nao conseguiu colocar na fila	*/
/*          TRUE  ==> se     conseguiu colocar na fila	*/
int poe( unsigned char byte)
{
	if (pin==pout)	return(FALSE);			/* fila cheia */
	fila[pin]=byte;
	if (++pin==TAM_FILA)	pin=0;
	return(TRUE);
}


/* Receber uma opcao numerica */
le_opc_num(int max)
{
	int	x;
	while(TRUE){
		scanf("%d",&x);
		if (x>0 && x<max+1)	break;
		printf("\nOpcoes de 1 ate %d. Tente novamente.",max);
	}
	return(x);
}

/* Receber uma opcao com letras */
le_opc(char *op)
{
	char	c[20],*x;
	while(TRUE){
		scanf("%s",c);
		x=strchr(op,c[0]);
		if ( x==NULL){
				printf("\nAs opcoes validas sao (%s)",op);
				printf("\nTente novamente : ");
		}
		else	break;
	}
	return(1+(x-op));
}

/* ler as teclas de F1 a F10 e ESC. Nao prende a execucao	*/
int getfuncao(void)
{
	union REGS regs;
	regs.h.ah=1;
	int86(0x16, &regs, &regs);
	if ((regs.x.flags&0x40) == 0x40)	regs.h.ah=0;
	else{
		regs.h.ah=0;
		int86(0x16, &regs, &regs);
		if (!((regs.h.al==0) && ((regs.h.ah>=59)&&(regs.h.ah<=68)) ) ){
			if (regs.h.ah==1 && regs.h.al==27)	regs.h.ah=27;
			else											regs.h.ah=0;
		}
	}
	return(regs.h.ah);
}

/* Imprimir uma possivel mensagem da placa de provas	*/
int rec_msg(void)
{
	unsigned char byte;
	printf("\n\nAguardando chegada de uma possivel mensagem serial\n");
	printf("Qualquer tecla para seguir adiante\n\n");
	while(~kbhit()){
		if (tira(&byte)==TRUE)	printf("%c",byte);
	}
	return(TRUE);
}

/* rotina para fazer o disassembly	*/
int disassembly(void){
	int i;
	unsigned int pc,data,lsn;
	int expr;
	int expr0,expr1,expr2,expr3,expr4,expr5,expr6,expr7;
	int expr8,expr9,expra,exprb,exprc,exprd,expre,exprf;

	pc=(reg[PCH]<<8) + reg[PCL];
	data=(reg[PC_1]<<8) + reg[PC_2];
	for (i=0; i<15; i++)		instr[i]='\0';

	/* ver faixa de valores do OP CODE	*/
	expr0 = reg[PC_0]>=0x04 && reg[PC_0]<=0xf;	/*  4 -->  F 	*/
	expr1 = reg[PC_0]>=0x14 && reg[PC_0]<=0x1f; 	/* 14 --> 1F	*/
	expr2 = reg[PC_0]>=0x24 && reg[PC_0]<=0x2f; 	/* 24 --> 2F	*/
	expr3 = reg[PC_0]>=0x34 && reg[PC_0]<=0x3f; 	/* 34 --> 3F	*/
	expr4 = reg[PC_0]>=0x42 && reg[PC_0]<=0x4f; 	/* 42 --> 4F	*/
	expr5 = reg[PC_0]>=0x52 && reg[PC_0]<=0x5f; 	/* 52 --> 5F	*/
	expr6 = reg[PC_0]>=0x62 && reg[PC_0]<=0x6f; 	/* 62 --> 6F	*/
	expr7 = reg[PC_0]>=0x74 && reg[PC_0]<=0x7f; 	/* 74 --> 7F	*/
	expr8 = reg[PC_0]>=0x85 && reg[PC_0]<=0x8f; 	/* 85 --> 8F	*/
	expr9 = reg[PC_0]>=0x94 && reg[PC_0]<=0x9f; 	/* 94 --> 9F	*/
	expra = reg[PC_0]>=0xa6 && reg[PC_0]<=0xaf; 	/* a6 --> aF	*/
	exprb = reg[PC_0]>=0xb4 && reg[PC_0]<=0xbf; 	/* b4 --> bF	*/
	exprc = reg[PC_0]>=0xc5 && reg[PC_0]<=0xcf; 	/* c5 --> cF	*/
	exprd = reg[PC_0]>=0xd8 && reg[PC_0]<=0xdf; 	/* d8 --> dF	*/
	expre = reg[PC_0]>=0xe5 && reg[PC_0]<=0xef; 	/* e5 --> eF	*/
	exprf = reg[PC_0]>=0xf5 && reg[PC_0]<=0xff; 	/* f5 --> fF	*/

	expr  =expr0||expr1||expr2||expr3||expr4||expr5||expr6||expr7;
	expr  =expr||expr8||expr9||expra||exprb||exprc||exprd||expre||exprf;


	/* usando as expr	*/
	if (expr){
		lsn=reg[PC_0]&0xf; 	/* separar nibbles	*/

		/* INC DEC 	*/
		if (expr0||expr1){
			if (expr0)			strcpy(instr,"INC ");
			else        		strcpy(instr,"DEC ");
			if 	  (lsn==4) 	strcpy(instr+strlen(instr),"A");
			else if (lsn==5)	qual_reg(reg[PC_1]);
			else					qual_rn(reg[PC_0]);
		}

		/* ADD, ADDC, SUBB, XCH, MOV	*/
		else if (expr2||expr3||expr9||exprc||expre){
			if      (expr2) 	strcpy(instr,"ADD A,");
			else if (expr3)	strcpy(instr,"ADDC A,");
			else if (expr9)	strcpy(instr,"SUBB A,");
			else if (exprc)	strcpy(instr,"XCH A,");
			else					strcpy(instr,"MOV A,");
			if (lsn==4){
				strcpy(instr+strlen(instr),"#");
				itoa(reg[PC_1],instr+strlen(instr),16);
			}
			else if (lsn==5) qual_reg(reg[PC_1]);
			else				  qual_rn(reg[PC_0]);
		}

		/* ORL, ANL, XRL	*/
		else if (expr4||expr5||expr6){
			if   	  (expr4)	strcpy(instr,"ORL ");
			else if (expr5)	strcpy(instr,"ANL ");
			else					strcpy(instr,"XRL ");
			if (lsn==2){		qual_reg(reg[PC_1]);
									strcpy(instr+strlen(instr),",A");	}
			else if (lsn==3){ qual_reg(reg[PC_1]);
									strcpy(instr+strlen(instr),",#");
									itoa(reg[PC_2],instr+strlen(instr),16);	}
			else if (lsn==4){ strcpy(instr+strlen(instr),"A,#");
									itoa(reg[PC_1],instr+strlen(instr),16);	}
			else if (lsn==5){ strcpy(instr+strlen(instr),"A,");
									itoa(reg[PC_1],instr+strlen(instr),16);	}
			else 				 { strcpy(instr+strlen(instr),"A,");
									qual_rn(reg[PC_0]);	}
		}

		/* MOV	*/
		else if (expr7||expr8||expra||exprf){
			strcpy(instr,"MOV ");
			if (expr7){
				if 	  (lsn==4)	strcpy(instr+strlen(instr),"A");
				else if (lsn==5)	qual_reg(reg[PC_1]);
				else					qual_rn(reg[PC_0]);
				strcpy(instr+strlen(instr),",#");
				itoa(reg[PC_1+(1&(lsn==5))],instr+strlen(instr),16);
			}
			else if (expr8){
				if (lsn==5){	qual_reg(reg[PC_2]);
									strcpy(instr+strlen(instr),",");
									qual_reg(reg[PC_1]);
				}
				else{				qual_reg(reg[PC_1]);
									strcpy(instr+strlen(instr),",");
									qual_rn(reg[PC_0]);
				}
			}
			else if (expra){
				qual_rn(reg[PC_0]);
				strcpy(instr+strlen(instr),",");
				qual_reg(reg[PC_1]);
			}
			else{
				if (lsn==5)	qual_reg(reg[PC_1]);
				else			qual_rn(reg[PC_0]);
				strcpy(instr+strlen(instr),",A");
			}
		}

		/* CJNE	*/
		else if (exprb){
			strcpy(instr,"CJNE ");
			if (lsn==4||lsn==5) strcpy(instr+strlen(instr),"A");
			else					  qual_rn(reg[PC_0]);
			strcpy(instr+strlen(instr),",");
			if (lsn==5)   qual_reg(reg[PC_1]);
			else{			  strcpy(instr+strlen(instr),"#");
							  itoa(reg[PC_1],instr+strlen(instr),16);
			}
			strcpy(instr+strlen(instr),",");
			itoa(relativo(pc,reg[PC_2]),instr+strlen(instr),16);
		}

		/* DJNZ	*/
		else{
			strcpy(instr,"DJNZ ");
			qual_rn(reg[PC_0]);
			strcpy(instr+strlen(instr),",");
			itoa(relativo(pc,reg[PC_1]),instr+strlen(instr),16);
		}
		return(TRUE);
	}
	/* termina uso das expr	*/

/*	printf("&0x1f = %04X\n",reg[PC_0]&0x1f);  */
	if ( (reg[PC_0]&0x1f) == 0x11){
		data=(reg[PC_0]>>5)*256+reg[PC_1];
		pc=(pc&0xf800)+data;
		strcpy(instr,"ACALL ");
		itoa(pc,instr+strlen(instr),16);
		return(TRUE);
	}

	if ( (reg[PC_0]&0x1f) == 1){
		data=(reg[PC_0]>>5)*256+reg[PC_1];
		pc=(pc&0xf800)+data;
		strcpy(instr,"AJMP ");
		itoa(pc,instr+strlen(instr),16);
		return(TRUE);
	}

	switch (reg[PC_0]){
		case 0x00:	strcpy(instr,"NOP");
						break;
		case 0x02:
		case 0x12: 	if (reg[PC_0]==2)	strcpy(instr,"LJMP ");
						else					strcpy(instr,"LCALL ");
						data=(reg[PC_1]<<8)+reg[PC_2];
						itoa(data,instr+strlen(instr),16);
						break;
		case 0x03:	strcpy(instr,"RR A");	break;

		case 0x10: 	strcpy(instr,"JBC ");
						qual_bit(reg[PC_1]);
						strcpy(instr+strlen(instr),",");
						itoa(relativo(pc,reg[PC_2]),instr+strlen(instr),16);
						break;
		case 0x13:	strcpy(instr,"RRC A");	break;
		case 0x20:	strcpy(instr,"JB ");
   					qual_bit(reg[PC_1]);
						strcpy(instr+strlen(instr),",");
						itoa(relativo(pc,reg[PC_2]),instr+strlen(instr),16);
						break;
		case 0x22:	strcpy(instr,"RET");	break;
		case 0x23:	strcpy(instr,"RL A");	break;
		case 0x30:	strcpy(instr,"JNB ");
   					qual_bit(reg[PC_1]);
						strcpy(instr+strlen(instr),",");
						itoa(relativo(pc,reg[PC_2]),instr+strlen(instr),16);
						break;
		case 0x32:	strcpy(instr,"RETI");	break;
		case 0x33:	strcpy(instr,"RLC A");	break;
		case 0x40:	strcpy(instr,"JC ");
						itoa(relativo(pc,reg[PC_1]),instr+strlen(instr),16);
						break;
		case 0x50:	strcpy(instr,"JNC ");
						itoa(relativo(pc,reg[PC_1]),instr+strlen(instr),16);
						break;
		case 0x60:	strcpy(instr,"JZ ");
						itoa(relativo(pc,reg[PC_1]),instr+strlen(instr),16);
						break;
		case 0x70:	strcpy(instr,"JNZ ");
						itoa(relativo(pc,reg[PC_1]),instr+strlen(instr),16);
						break;
		case 0x72:	strcpy(instr,"ORL C,");
						qual_bit(reg[PC_1]);
						break;
		case 0x73:	strcpy(instr,"JMP @A+DPTR");	break;
		case 0x80:	strcpy(instr,"SJMP ");
						itoa(relativo(pc,reg[PC_1]),instr+strlen(instr),16);
						break;
		case 0x82:	strcpy(instr,"ANL C,");
						qual_bit(reg[PC_1]);
						break;
		case 0x83:	strcpy(instr,"MOVC A,@A+PC");	break;
		case 0x84:	strcpy(instr,"DIV AB");	break;
		case 0x90:	strcpy(instr,"MOV DPTR,#");
						itoa(data,instr+strlen(instr),16);
						break;
		case 0x92:	strcpy(instr,"MOV ");
						qual_bit(reg[PC_1]);
						strcpy(instr+strlen(instr),",C");
						break;
		case 0x93:	strcpy(instr,"MOVC A,@A+DPTR");	break;
		case 0xA0:	strcpy(instr,"ORL C,/");
   					qual_bit(reg[PC_1]);
						break;
		case 0xA2:	strcpy(instr,"MOV C,");
						qual_bit(reg[PC_1]);
						break;
		case 0xA3:	strcpy(instr,"INC DPTR");	break;
		case 0xA4:	strcpy(instr,"MUL AB");	break;
		case 0xA5:	strcpy(instr,"??????");	break;
		case 0xB0:	strcpy(instr,"ANL C,/");
   					qual_bit(reg[PC_1]);
						break;
		case 0xB2:	strcpy(instr,"CPL ");
						qual_bit(reg[PC_1]);
						break;
		case 0xB3:	strcpy(instr,"CPL C");	break;
		case 0xC0:	strcpy(instr,"PUSH ");
						qual_reg(reg[PC_1]);
						break;
		case 0xC2:	strcpy(instr,"CLR ");
						qual_bit(reg[PC_1]);
						break;
		case 0xC3:	strcpy(instr,"CLR C");	break;
		case 0xC4:	strcpy(instr,"SWAP A");	break;
		case 0xD0:	strcpy(instr,"POP ");
						qual_reg(reg[PC_1]);
						break;
		case 0xD2:	strcpy(instr,"SETB ");
						qual_bit(reg[PC_1]);
						break;
		case 0xD3:	strcpy(instr,"SETB C");	break;
		case 0xD4:	strcpy(instr,"DA A");	break;
		case 0xD5:	strcpy(instr,"DJNZ ");
						qual_reg(reg[PC_1]);
						strcpy(instr+strlen(instr),",");
						itoa(relativo(pc,reg[PC_2]),instr+strlen(instr),16);
						break;
		case 0xD6:
		case 0xD7:	strcpy(instr,"XCHD A,");
						qual_rn(reg[PC_0]);
						break;
		case 0xE0:	strcpy(instr,"MOVX A,@DPTR");	break;
		case 0xE2:
		case 0xE3:	strcpy(instr,"MOVX A,");
						qual_rn(reg[PC_0]);
						break;
		case 0xE4:	strcpy(instr,"CLR A");	break;
		case 0xF0:	strcpy(instr,"MOVX @DPTR,A");
						break;
		case 0xF2:
		case 0xF3:	strcpy(instr,"MOVX ");
						qual_rn(reg[PC_0]);
						strcpy(instr+strlen(instr),",A");
						break;
		case 0xF4:	strcpy(instr,"CPL A");	break;


	}
	return(TRUE);
}

qual_rn(unsigned char dado)
{
	unsigned char x;
	x=dado&0xf;
	if 	  (x==6)	strcpy(instr+strlen(instr),"@R0");
	else if (x==7)	strcpy(instr+strlen(instr),"@R1");
	else{					strcpy(instr+strlen(instr),"R");
							x-=8;
							itoa(x,instr+strlen(instr),16);
	}
	return(TRUE);
}

int qual_reg (unsigned char rg)
{
	int flag=FALSE;
	switch (rg){
		case 0x80:	strcpy(instr+strlen(instr),"P0");	flag=TRUE;   break;
		case 0x81:	strcpy(instr+strlen(instr),"SP");	flag=TRUE;   break;
		case 0x82:	strcpy(instr+strlen(instr),"DPL");	flag=TRUE;   break;
		case 0x83:	strcpy(instr+strlen(instr),"DPH");	flag=TRUE;   break;
		case 0x87:	strcpy(instr+strlen(instr),"PCON");	flag=TRUE;   break;
		case 0x88:	strcpy(instr+strlen(instr),"TCON");	flag=TRUE;   break;
		case 0x89:	strcpy(instr+strlen(instr),"TMOD");	flag=TRUE;   break;
		case 0x8A:	strcpy(instr+strlen(instr),"TL0");	flag=TRUE;   break;
		case 0x8B:	strcpy(instr+strlen(instr),"TH0");	flag=TRUE;   break;
		case 0x8C:	strcpy(instr+strlen(instr),"TL1");	flag=TRUE;   break;
		case 0x8D:	strcpy(instr+strlen(instr),"TH1");	flag=TRUE;   break;
		case 0x90:	strcpy(instr+strlen(instr),"P1");	flag=TRUE;   break;
		case 0x98:	strcpy(instr+strlen(instr),"SCON");	flag=TRUE;   break;
		case 0x99:	strcpy(instr+strlen(instr),"SBUF");	flag=TRUE;   break;
		case 0xA0:	strcpy(instr+strlen(instr),"P2");	flag=TRUE;   break;
		case 0xA8:	strcpy(instr+strlen(instr),"IE");	flag=TRUE;   break;
		case 0xB0:	strcpy(instr+strlen(instr),"P3");	flag=TRUE;   break;
		case 0xB8:	strcpy(instr+strlen(instr),"IP");	flag=TRUE;   break;
		case 0xD0:	strcpy(instr+strlen(instr),"PSW");	flag=TRUE;   break;
		case 0xE0:	strcpy(instr+strlen(instr),"ACC");	flag=TRUE;   break;
		case 0xF0:	strcpy(instr+strlen(instr),"B");	flag=TRUE;   break;
	}
	if (flag)	return(TRUE);
	else{
		itoa(rg,instr+strlen(instr),16);
		return(TRUE);
	}
}


int qual_bit (unsigned char bit)
{
	int expr1,expr2,expr3,expr4,expr5,expr6;
/*	printf("busca bit %02X\n",bit); */

	if (bit<127){
		itoa(bit&7,instr+strlen(instr),10);
		return(TRUE);
	}

	expr1= 0x80<=bit && bit<=0x87;	/* P0	*/
	expr2= 0x90<=bit && bit<=0x97;	/* P1	*/
	expr3= 0xa0<=bit && bit<=0xa7;	/* P2	*/
	expr4= 0xb0<=bit && bit<=0xb7;	/* P3	*/
	expr5= 0xe0<=bit && bit<=0xe7;	/* A 	*/
	expr6= 0xf0<=bit && bit<=0xf7;	/* B 	*/
	if (expr1 || expr2 || expr3 || expr4 || expr5 || expr6) {
		if (expr1){
			strcpy(instr+strlen(instr),"P0.");
			itoa(bit&7,instr+strlen(instr),10);
		}
		else if (expr2){
			strcpy(instr+strlen(instr),"P1.");
			itoa(bit&7,instr+strlen(instr),10);
		}
		else if (expr3){
			strcpy(instr+strlen(instr),"P2.");
			itoa(bit&7,instr+strlen(instr),10);
		}
		else if (expr4){
			strcpy(instr+strlen(instr),"P3.");
			itoa(bit&7,instr+strlen(instr),10);
		}
		else if (expr5){
			strcpy(instr+strlen(instr),"Acc.");
			itoa(bit&7,instr+strlen(instr),10);
		}
		else {
			strcpy(instr+strlen(instr),"B.");
			itoa(bit&7,instr+strlen(instr),10);
		}
		return(TRUE);
	}

	switch (bit){

		case 0x88:	strcpy(instr+strlen(instr),"IT0");	break;
		case 0x89:	strcpy(instr+strlen(instr),"IE0");	break;
		case 0x8A:	strcpy(instr+strlen(instr),"IT1");	break;
		case 0x8B:	strcpy(instr+strlen(instr),"IE1");	break;
		case 0x8C:	strcpy(instr+strlen(instr),"TR0");	break;
		case 0x8D:	strcpy(instr+strlen(instr),"TF0");	break;
		case 0x8E:	strcpy(instr+strlen(instr),"TR1");	break;
		case 0x8F:	strcpy(instr+strlen(instr),"TF1");	break;

		case 0x98:	strcpy(instr+strlen(instr),"RI");	break;
		case 0x99:	strcpy(instr+strlen(instr),"TI");	break;
		case 0x9A:	strcpy(instr+strlen(instr),"RB8");	break;
		case 0x9B:	strcpy(instr+strlen(instr),"TB8");	break;
		case 0x9C:	strcpy(instr+strlen(instr),"REN");	break;
		case 0x9D:	strcpy(instr+strlen(instr),"SM2");	break;
		case 0x9E:	strcpy(instr+strlen(instr),"SM1");	break;
		case 0x9F:	strcpy(instr+strlen(instr),"SM0");	break;

		case 0xA8:	strcpy(instr+strlen(instr),"EX0");	break;
		case 0xA9:	strcpy(instr+strlen(instr),"ET0");	break;
		case 0xAA:	strcpy(instr+strlen(instr),"EX1");	break;
		case 0xAB:	strcpy(instr+strlen(instr),"ET1");	break;
		case 0xAC:	strcpy(instr+strlen(instr),"ES");	break;
		case 0xAD:	strcpy(instr+strlen(instr),"???");	break;
		case 0xAE:	strcpy(instr+strlen(instr),"???");	break;
		case 0xAF:	strcpy(instr+strlen(instr),"EA");	break;

		case 0xB8:	strcpy(instr+strlen(instr),"PX0");	break;
		case 0xB9:	strcpy(instr+strlen(instr),"PT0");	break;
		case 0xBA:	strcpy(instr+strlen(instr),"PX1");	break;
		case 0xBB:	strcpy(instr+strlen(instr),"PT1");	break;
		case 0xBC:	strcpy(instr+strlen(instr),"PS");	break;
		case 0xBD:	strcpy(instr+strlen(instr),"???");	break;
		case 0xBE:	strcpy(instr+strlen(instr),"???");	break;
		case 0xBF:	strcpy(instr+strlen(instr),"???");	break;

		case 0xD0:	strcpy(instr+strlen(instr),"P");	break;
		case 0xD1:	strcpy(instr+strlen(instr),"???");	break;
		case 0xD2:	strcpy(instr+strlen(instr),"OV");	break;
		case 0xD3:	strcpy(instr+strlen(instr),"RS0");	break;
		case 0xD4:	strcpy(instr+strlen(instr),"RS1");	break;
		case 0xD5:	strcpy(instr+strlen(instr),"F0");	break;
		case 0xD6:	strcpy(instr+strlen(instr),"AC");	break;
		case 0xD7:	strcpy(instr+strlen(instr),"CY");	break;
	}
	return(TRUE);
}


int relativo (unsigned int pc, unsigned char rel)
{
	int	x;
	x=rel;
	if (rel>0x7f)	x=rel-0x100;
	return(pc+2+x);
}



/* rotina de help	*/
void help(void)
{
	int i,x,y,opc;
	x=wherex();
	y=wherey();
	guarda_tela(XWI,YWI,XWF,YWF);
	faz_moldura(XWI,YWI,XWF,YWF,texto);
	window(XWI,YWI,XWF,YWF);
	while(TRUE){
		clrscr();
		cprintf("          OPCOES :\n\r");
		cprintf("          A - Ajuda\n\r");
		cprintf("          C - Conselho\n\r");
		cprintf("          S - Sair\n\r");
		cprintf("          Selecione uma opcao : ");
		opc=le_opc_help("ACS");
		if (opc==3)	break;
		if (opc==1){
			clrscr();
			cprintf("\n\r  Voce sabe o que o sabio viu,");
			cprintf("\n\r  apos subir a montanha ? (Sim/Nao) ");
			i=le_opc_help("SN");
			if (i==1){
				cprintf("\n\n\r Bem. Entao nao vou contar.\n\n\r");
				qq_tecla_help();
			}
			else{
				cprintf("\n\n\r Ele vio o outro lado da montanha.\n\n\r");
				qq_tecla_help();
			}
		}
		if (opc==2){
			clrscr();
			cprintf("\n\r            PALAVRAS SABIAS");
			cprintf("\n\n\r   A vida e' como uma cebola !");
			cprintf("\n\r   Se voce a analisa camada apos camada,");
			cprintf("\n\r   nao chega a nada.\n\r\n");
			qq_tecla_help();
		}
	}
	window(1,1,80,25);
	recupera_tela(XWI,YWI,XWF,YWF);
	gotoxy(x,y);
}

/* esperar uma tecla qualquer */
int qq_tecla_help(void)
{
	int x;
	getch();
	cprintf("\n\r Uma tecla qualquer para continuar.\n\r");
	x=getch();
	while(x==CR || x==SPACE){
		cprintf("\n\r <<<<< Por que sera' que quando peco uma tecla,");
		cprintf("\n\r sempre  usam <ENTER> ou <SPACE> ? >>>>>\n\r");
		cprintf(" Ja' estou cansado !\n\r");
		cprintf(" Estas eu nao aceito. Tente outra. ");
		x=getch();
		printf("\n\r");
	}

	return(TRUE);
}

/* Receber uma opcao com letras para help */
le_opc_help(char *op)
{
	char	c[20],*x;
	while(TRUE){
		cscanf("%s",c);
		cprintf("\n\r");
		strupr(c);
		if (strlen(c)!=1){
			cprintf("---> Voce nao e' capaz de digitar corretamente ?\n\r");
			cprintf("---> Somente uma letra !\n\r");
		}
		x=strchr(op,c[0]);
		if ( x==NULL){
				cprintf("---> Ai caramba! Faca a selecao correta.\n\r");
				cprintf("\n\r---> As opcoes validas sao (%s).\n\r",op);
				cprintf("---> Tente novamente : ");
		}
		else	break;
	}
	return(1+(x-op));
}


faz_moldura(int xi, int yi, int xf, int yf, char *msg)
{
	int i,x;
	gotoxy(xi-1,yi-1);
	putch(SUP_ESQ);
	for (i=xi; i<=xf; i++)	putch(HOR);
	putch(SUP_DIR);
	for (i=yi; i<=yf; i++){
		gotoxy(xi-1,i);	putch(VER);
		gotoxy(xf+1,i);	putch(VER);
	}
	gotoxy(xi-1,yf+1);	putch(INF_ESQ);
	for (i=xi; i<=xf; i++)	putch(HOR);
	putch(INF_DIR);
	x=((xf-xi+3)-strlen(texto))/2;
	gotoxy(xi+x,yi-1);
	puts(msg);
	return(1);
}

guarda_tela(int xi, int yi, int xf, int yf)
{
	int dx,dy,x,y,des,w;
	dx=xf-xi+3;
	dy=yf-yi+3;
	w=0;
	for(y=0; y<dy; y++){
		for(x=0; x<dx; x++){
			des=(yi-2+y)*160+2*(xi-2+x);
			buffer[w++]=*(tela+des);
		}
	}
	return(1);
}

recupera_tela(int xi, int yi, int xf, int yf)
{
	int dx,dy,x,y,des,w;
	dx=xf-xi+3;
	dy=yf-yi+3;
	w=0;
	for(y=0; y<dy; y++){
		for(x=0; x<dx; x++){
			des=(yi-2+y)*160+2*(xi-2+x);
			*(tela+des)=buffer[w++];
		}
	}
	return(1);
}

/* Funcao para aguardar que se tecle <ENTER>	*/
void get_enter(void)
{
	int c=0;
	printf("Tecle <ENTER> para continuar. ");
	while(c!='\r')		c=getch();
}

/* Funcao para aguardar que se tecle <ENTER>	*/
/* Nao coloca nenhuma mensagem					*/
void get_enter_silencio(void)
{
	int c=0;
	while(c!='\r')		c=getch();
}
